From 5f3a4a0ca40f5d222596d384c4dd682fcc8a8614 Mon Sep 17 00:00:00 2001
From: Shveta Malik <shveta.malik@gmail.com>
Date: Wed, 6 Dec 2023 15:58:21 +0530
Subject: [PATCH v43 3/3] Allow slot-sync worker to wait for the cascading
 standbys.

The GUC standby_slot_names is needed to be set on first standby
in order to allow it to wait for confirmation for cascading
standbys before updating logical 'synced' slots in slot-sync worker.
The intent is that the logical slots (synced ones) should not go
ahead of cascading standbys.

For the user created slots on first standby, we already have this wait
logic in place in logical walsender and in pg_logical_slot_get_changes_guts(),
but for synced slots (which can not be consumed yet), we need to make
sure that they are not going ahead of cascading standbys and that is
acheived by introducing the wait in slot-sync worker before we actually
update the slots.
---
 src/backend/replication/logical/slotsync.c | 89 ++++++++++++++++++++--
 src/backend/replication/walsender.c        |  4 +-
 src/include/replication/walsender.h        |  5 ++
 3 files changed, 88 insertions(+), 10 deletions(-)

diff --git a/src/backend/replication/logical/slotsync.c b/src/backend/replication/logical/slotsync.c
index 8266ebe537..a563b4bb25 100644
--- a/src/backend/replication/logical/slotsync.c
+++ b/src/backend/replication/logical/slotsync.c
@@ -108,7 +108,9 @@ static TimestampTz last_update_time;
  */
 #define WORKER_PRIMARY_CATCHUP_WAIT_ATTEMPTS 5
 
-static void ProcessSlotSyncInterrupts(WalReceiverConn *wrconn);
+static void ProcessSlotSyncInterrupts(WalReceiverConn *wrconn,
+									  List **standby_slots);
+
 
 /*
  * Wait for remote slot to pass locally reserved position.
@@ -164,7 +166,7 @@ wait_for_primary_slot_catchup(WalReceiverConn *wrconn, RemoteSlot *remote_slot,
 		CHECK_FOR_INTERRUPTS();
 
 		/* Handle any termination request if any */
-		ProcessSlotSyncInterrupts(wrconn);
+		ProcessSlotSyncInterrupts(wrconn, NULL /* standby_slots */ );
 
 		res = walrcv_exec(wrconn, cmd.data, WAIT_OUTPUT_COLUMN_COUNT, slotRow);
 
@@ -480,6 +482,52 @@ construct_slot_query(StringInfo s)
 					 " WHERE failover and sync_state != 'i'");
 }
 
+/*
+ * Wait for cascading physical standbys corresponding to physical slots
+ * specified in standby_slot_names GUC to confirm receiving given lsn.
+ */
+static void
+wait_for_standby_confirmation(XLogRecPtr wait_for_lsn,
+							  WalReceiverConn *wrconn)
+{
+	List	   *standby_slots;
+
+	/* Nothing to be done */
+	if (strcmp(standby_slot_names, "") == 0)
+		return;
+
+	standby_slots = GetStandbySlotList(true);
+
+	for (;;)
+	{
+		int			rc;
+
+		WalSndFilterStandbySlots(wait_for_lsn, &standby_slots);
+
+		/* Exit if done waiting for every slot. */
+		if (standby_slots == NIL)
+			break;
+
+		/*
+		 * This will reload configuration and will refresh the standby_slots
+		 * as well provided standby_slot_names GUC is changed by the user.
+		 */
+		ProcessSlotSyncInterrupts(wrconn, &standby_slots);
+
+		/*
+		 * XXX: Is waiting for 5 second before retrying enough or more or
+		 * less?
+		 */
+		rc = WaitLatch(MyLatch,
+					   WL_LATCH_SET | WL_TIMEOUT | WL_EXIT_ON_PM_DEATH,
+					   5000L,
+					   WAIT_EVENT_WAL_SENDER_WAIT_FOR_STANDBY_CONFIRMATION);
+
+		if (rc & WL_LATCH_SET)
+			ResetLatch(MyLatch);
+	}
+}
+
 /*
  * Synchronize single slot to given position.
  *
@@ -760,6 +808,7 @@ synchronize_slots(WalReceiverConn *wrconn)
 	List	   *remote_slot_list = NIL;
 	MemoryContext oldctx = CurrentMemoryContext;
 	long		naptime = WORKER_DEFAULT_NAPTIME_MS;
+	XLogRecPtr	max_confirmed_lsn = 0;
 	ListCell   *cell;
 	bool		slot_updated = false;
 	TimestampTz now;
@@ -817,6 +866,9 @@ synchronize_slots(WalReceiverConn *wrconn)
 		remote_slot->plugin = TextDatumGetCString(slot_getattr(slot, 2, &isnull));
 		Assert(!isnull);
 
+		if (remote_slot->confirmed_lsn > max_confirmed_lsn)
+			max_confirmed_lsn = remote_slot->confirmed_lsn;
+
 		/*
 		 * It is possible to get null values for LSN and Xmin if slot is
 		 * invalidated on the primary server, so handle accordingly.
@@ -860,6 +912,17 @@ synchronize_slots(WalReceiverConn *wrconn)
 	 */
 	drop_obsolete_slots(remote_slot_list);
 
+	/*
+	 * If there are cascading standbys, wait for their confirmation before we
+	 * update synced logical slots locally.
+	 *
+	 * Instead of waiting on confirmation for lsn of each slot, let us wait
+	 * once for confirmation on max_confirmed_lsn. If that is confirmed by
+	 * each cascading standby, we are good to update all the slots.
+	 */
+	if (remote_slot_list)
+		wait_for_standby_confirmation(max_confirmed_lsn, wrconn);
+
 	/* Now sync the slots locally */
 	foreach(cell, remote_slot_list)
 	{
@@ -1035,7 +1098,7 @@ validate_slotsync_parameters(char **dbname)
  * if validaity fails.
  */
 static void
-slotsync_reread_config(WalReceiverConn *wrconn)
+slotsync_reread_config(WalReceiverConn *wrconn, List **standby_slots)
 {
 	char	   *conninfo = pstrdup(PrimaryConnInfo);
 	char	   *slotname = pstrdup(PrimarySlotName);
@@ -1049,8 +1112,14 @@ slotsync_reread_config(WalReceiverConn *wrconn)
 	dbname = walrcv_get_dbname_from_conninfo(PrimaryConnInfo);
 	Assert(dbname);
 
-	ConfigReloadPending = false;
-	ProcessConfigFile(PGC_SIGHUP);
+	/*
+	 * Reload configs and recreate the standby_slot_names_list if GUC
+	 * standby_slot_names changed.
+	 */
+	if (standby_slots)
+		WalSndRereadConfigAndReInitSlotList(standby_slots);
+	else
+		ProcessConfigFile(PGC_SIGHUP);
 
 	conninfoChanged = strcmp(conninfo, PrimaryConnInfo) != 0;
 	slotnameChanged = strcmp(slotname, PrimarySlotName) != 0;
@@ -1093,7 +1162,8 @@ slotsync_reread_config(WalReceiverConn *wrconn)
  * Interrupt handler for main loop of slot sync worker.
  */
 static void
-ProcessSlotSyncInterrupts(WalReceiverConn *wrconn)
+ProcessSlotSyncInterrupts(WalReceiverConn *wrconn,
+						  List **standby_slots)
 {
 	CHECK_FOR_INTERRUPTS();
 
@@ -1109,7 +1179,10 @@ ProcessSlotSyncInterrupts(WalReceiverConn *wrconn)
 
 
 	if (ConfigReloadPending)
-		slotsync_reread_config(wrconn);
+	{
+		ConfigReloadPending = false;
+		slotsync_reread_config(wrconn, standby_slots);
+	}
 }
 
 /*
@@ -1183,7 +1256,7 @@ ReplSlotSyncWorkerMain(Datum main_arg)
 		int			rc;
 		long		naptime;
 
-		ProcessSlotSyncInterrupts(wrconn);
+		ProcessSlotSyncInterrupts(wrconn, NULL /* standby_slots */ );
 
 		naptime = synchronize_slots(wrconn);
 
diff --git a/src/backend/replication/walsender.c b/src/backend/replication/walsender.c
index 7b6962f354..537b5e27dc 100644
--- a/src/backend/replication/walsender.c
+++ b/src/backend/replication/walsender.c
@@ -1609,7 +1609,7 @@ PhysicalWakeupLogicalWalSnd(void)
  * Reload the config file and reinitialize the standby slot list if the GUC
  * standby_slot_names has changed.
  */
-static void
+void
 WalSndRereadConfigAndReInitSlotList(List **standby_slots)
 {
 	char	   *pre_standby_slot_names = pstrdup(standby_slot_names);
@@ -1634,7 +1634,7 @@ WalSndRereadConfigAndReInitSlotList(List **standby_slots)
  * it removes slots that have been invalidated, dropped, or converted to
  * logical slots.
  */
-static void
+void
 WalSndFilterStandbySlots(XLogRecPtr wait_for_lsn, List **standby_slots)
 {
 	ListCell   *lc;
diff --git a/src/include/replication/walsender.h b/src/include/replication/walsender.h
index ecd212c7b6..07b2c719ae 100644
--- a/src/include/replication/walsender.h
+++ b/src/include/replication/walsender.h
@@ -13,6 +13,7 @@
 #define _WALSENDER_H
 
 #include "access/xlogdefs.h"
+#include "nodes/pg_list.h"
 
 /*
  * What to do with a snapshot in create replication slot command.
@@ -48,7 +49,11 @@ extern void WalSndWaitStopping(void);
 extern void HandleWalSndInitStopping(void);
 extern void WalSndRqstFileReload(void);
 extern void PhysicalWakeupLogicalWalSnd(void);
+extern void WalSndFilterStandbySlots(XLogRecPtr wait_for_lsn,
+									 List **standby_slots);
 extern void WalSndWaitForStandbyConfirmation(XLogRecPtr wait_for_lsn);
+extern List *WalSndGetStandbySlots(void);
+extern void WalSndRereadConfigAndReInitSlotList(List **standby_slots);
 
 /*
  * Remember that we want to wakeup walsenders later
-- 
2.34.1

